/* **************************************************************************
 *    Copyright (C) 2005 by Rob Brown                                       *
 *    rob@cobbleware.com                                                    *
 *                                                                          *
 *    C Implementation: sdk2x_sd_part                                       *
 *                                                                          *
 *    This program is free software; you can redistribute it and/or modify  *
 *    it under the terms of the GNU General Public License as published by  *
 *    the Free Software Foundation; either version 2 of the License, or     *
 *    (at your option) any later version.                                   *
 *                                                                          *
 *    This program is distributed in the hope that it will be useful,       *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *    GNU General Public License for more details.                          *
 *                                                                          *
 *    You should have received a copy of the GNU General Public License     *
 *    along with this program; if not, write to the                         *
 *    Free Software Foundation, Inc.,                                       *
 *    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 * **************************************************************************/

#include "sdk2x_sd.h"
#include "sdk2x_priv.h"
#include "sdk2x_sd_part.h"
#include "sdk2x_sd_hw.h"

#include "stdlib.h"

#if (SDK2X_SD_DEBUG > 1)
#include "stdio.h"
#endif

#if (SDK2X_SD_DEBUG > 0)
void DumpSector (void *p);

const char FatNames[3][6] = { "FAT12", "FAT16", "FAT32" };
#endif

typedef struct _tagSdPartInternal
{
	unsigned char *pPartitionTable;
	unsigned int nPartTableSize;
	unsigned int nBlocksPer512Bytes;
} SdPartInternal_t;

typedef struct _tagPartition
{
	/* Partition table details */
#if (SDK2X_SD_DEBUG > 0)
	int nDriveNum;
	int nStartCylinder;
	int nStartHead;
	int nStartSector;
	int nEndCylinder;
	int nEndHead;
	int nEndSector;
#endif
	int nType;
	int nStartLBA;
	int nSize;

	/* BPB details */
	unsigned char OEMName[9];
	unsigned int nBytesPerSector;
	unsigned int nSectorsPerCluster;
	unsigned int nReservedSectors;
	unsigned int nNumFATs;
	unsigned int nRootEntCnt;
	unsigned int nTotSec16;
	unsigned int nMedia;
	unsigned int nFATsize16;
	unsigned int nSectorsPerTrack;
	unsigned int nNumHeads;
	unsigned int nHiddenSectors;
	unsigned int nTotSec32;

	/* These are present for FAT12/FAT16 and for FAT32, but on FAT32 they are at a different place in the BPB. */
	unsigned int nDriveNumber;
	unsigned int nBootSig;
	unsigned int nVolID;
	unsigned char VolLab[12];
	unsigned char FileSysType[9];

	/* Only for FAT32 */
	unsigned int nFATsize32;
	unsigned int nExtFlags;
	unsigned int nFSVer;
	unsigned int nRootClus;
	unsigned int nFSInfo;
	unsigned int nBkBootSec;

	/* Computed values */
	unsigned int nFATSize;
	unsigned int nTotalSectors;
	unsigned int nRootDirSectors;
	unsigned int nFirstDataSector;
	unsigned int nDataSectors;
	unsigned int nCountOfClusters;
	unsigned int nFATType;
} Partition_t;

static SdPartInternal_t SdPartInternal;
static Partition_t Partitions[4];

int bad_partition;

static unsigned int GetInt (unsigned char *pBuffer, int offset, int numBytes)
{
	int i,r;

	r=0;

	for (i=0;i<numBytes;i++)
		r += ( ((unsigned int)pBuffer[offset + i]) << (8 * i));

	return r;
}

static void GetString (unsigned char *pSource, int offset, unsigned char *pDest, int length)
{
	int i;
	unsigned char c;

	for (i=0;i<length;i++)
	{
		c = pSource[i + offset];
		if (!c)
			c = ' ';
		if (c<32)
			c = '.';
		pDest[i] = c;
	}

	pDest[length] = '\0';
}

static int LoadPartitionBootRecord (int nPartition)
{
	unsigned char tempbuf[1024];
	int r;

	r = sdk2x_SdReadSectors(Partitions[nPartition].nStartLBA * sdk2x_SdCardInfo.BlockSize, tempbuf, SdPartInternal.nBlocksPer512Bytes);

	if (r)
	{
#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "SdReadSector returned %d\r\n",r);
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif
		return SDK2X_ERR_SDPT_MBR_LOAD_FAIL;
	}

#if (SDK2X_SD_DEBUG > 2)
	DumpSector (tempbuf);
#endif

	if ((tempbuf[510]!=0x55)||(tempbuf[511]!=0xAA))
	{
#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "Partition %d MBR read incorrect! Key = 0x%04X\r\n",nPartition,*(unsigned short*)&(tempbuf[510]));
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif
		return SDK2X_ERR_SDPT_PTABLE_CORRUPT;
	}

	GetString(tempbuf, 3, Partitions[nPartition].OEMName, 8);
	Partitions[nPartition].nBytesPerSector = GetInt (tempbuf, 11, 2);
	Partitions[nPartition].nSectorsPerCluster = GetInt (tempbuf, 13, 1);
	Partitions[nPartition].nReservedSectors = GetInt (tempbuf, 14, 2);
	Partitions[nPartition].nNumFATs = GetInt (tempbuf, 16, 1);
	Partitions[nPartition].nRootEntCnt = GetInt (tempbuf, 17, 2);
	Partitions[nPartition].nTotSec16 = GetInt (tempbuf, 19, 2);
	Partitions[nPartition].nMedia = GetInt (tempbuf, 21, 1);
	Partitions[nPartition].nFATsize16 = GetInt (tempbuf, 22, 2);
	Partitions[nPartition].nSectorsPerTrack = GetInt (tempbuf, 24, 2);
	Partitions[nPartition].nNumHeads = GetInt (tempbuf, 26, 2);
	Partitions[nPartition].nHiddenSectors = GetInt (tempbuf, 28, 4);
	Partitions[nPartition].nTotSec32 = GetInt (tempbuf, 32, 4);

	/* TODO: If nTotSec16 and nTotalSectors are both non-zero, throw an error or something
	         although this is how the MS FAT spec says to do it */
	if (Partitions[nPartition].nTotSec16)
		Partitions[nPartition].nTotalSectors = Partitions[nPartition].nTotSec16;
	else
		Partitions[nPartition].nTotalSectors = Partitions[nPartition].nTotSec32;

	Partitions[nPartition].nDriveNumber = GetInt (tempbuf, 36, 1);
	Partitions[nPartition].nBootSig = GetInt (tempbuf, 38, 1);
	Partitions[nPartition].nVolID = GetInt (tempbuf, 39, 4);
	GetString(tempbuf, 43, Partitions[nPartition].VolLab, 11);
	GetString(tempbuf, 54, Partitions[nPartition].FileSysType, 8);

	Partitions[nPartition].nFATsize32 = GetInt (tempbuf, 36, 4);

	if (Partitions[nPartition].nFATsize16)
		Partitions[nPartition].nFATSize = Partitions[nPartition].nFATsize16;
	else
		Partitions[nPartition].nFATSize = Partitions[nPartition].nFATsize32;

#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "\r\nPartition %d information:\r\n\r\nOEM Name: \"%s\"  BPS: %d  SPC: %d  RsvdSec: %d\r\n",nPartition,
					Partitions[nPartition].OEMName, Partitions[nPartition].nBytesPerSector,
					Partitions[nPartition].nSectorsPerCluster, Partitions[nPartition].nReservedSectors);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "NumFATs: %d  RootEnts: %d  TotalSectors: %d  Media: %02X\r\n",
					Partitions[nPartition].nNumFATs, Partitions[nPartition].nRootEntCnt,
					Partitions[nPartition].nTotalSectors, Partitions[nPartition].nMedia);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "FAT Size: %d  SPT: %d  Heads: %d  HiddenSecs: %d\r\n",
					Partitions[nPartition].nFATSize, Partitions[nPartition].nSectorsPerTrack,
					Partitions[nPartition].nNumHeads, Partitions[nPartition].nHiddenSectors);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "Volume Label = \"%s\"  FAT Type string = \"%s\"\r\n",
					Partitions[nPartition].VolLab, Partitions[nPartition].FileSysType);
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif

	Partitions[nPartition].nRootDirSectors =
		((Partitions[nPartition].nRootEntCnt << 5) + (Partitions[nPartition].nBytesPerSector - 1)) / Partitions[nPartition].nBytesPerSector;

	Partitions[nPartition].nFirstDataSector =
			Partitions[nPartition].nReservedSectors + (Partitions[nPartition].nNumFATs * Partitions[nPartition].nFATSize) +
			Partitions[nPartition].nRootDirSectors;

	Partitions[nPartition].nDataSectors = Partitions[nPartition].nTotalSectors - Partitions[nPartition].nFirstDataSector;

	Partitions[nPartition].nCountOfClusters = Partitions[nPartition].nDataSectors / Partitions[nPartition].nSectorsPerCluster;

	if (Partitions[nPartition].nCountOfClusters < 4085)
		Partitions[nPartition].nFATType = SDK2X_FAT12;
	else if (Partitions[nPartition].nCountOfClusters < 65525)
		Partitions[nPartition].nFATType = SDK2X_FAT16;
	else
		Partitions[nPartition].nFATType = SDK2X_FAT32;

#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "\r\nRoot Directory Sectors = %d\r\n",Partitions[nPartition].nRootDirSectors);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "First Data Sector = %d\r\n",Partitions[nPartition].nFirstDataSector);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "Total Data Sectors = %d\r\n",Partitions[nPartition].nDataSectors);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "Count Of Clusters = %d\r\n",Partitions[nPartition].nCountOfClusters);
		sdk2x_UartSendString(0,sdk2x_private_string);

		sprintf (sdk2x_private_string, "Type of FAT is %s\r\n",FatNames[Partitions[nPartition].nFATType]);
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif

	return 0;
}

int sdk2x_SdPartInit (void)
{
	int r,i;

	r = sdk2x_SdHwInit();

	if (r)
		return r;

	SdPartInternal.nBlocksPer512Bytes = (sdk2x_SdCardInfo.BlockSize + 511) / sdk2x_SdCardInfo.BlockSize;
	SdPartInternal.nPartTableSize = (sdk2x_SdCardInfo.BlockSize * SdPartInternal.nBlocksPer512Bytes) + 4;

	SdPartInternal.pPartitionTable = (unsigned char *)malloc(SdPartInternal.nPartTableSize);
 	if (!SdPartInternal.pPartitionTable)
 		return SDK2X_ERR_SDPT_MALLOC;

	r = sdk2x_SdReadSectors(0, SdPartInternal.pPartitionTable, SdPartInternal.nBlocksPer512Bytes);

	if (r)
	{
#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "SdReadSector returned %d\r\n",r);
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif
		return SDK2X_ERR_SDPT_READ_PTABLE;
	}

	if ((SdPartInternal.pPartitionTable[510]!=0x55)||(SdPartInternal.pPartitionTable[511]!=0xAA))
	{
#if (SDK2X_SD_DEBUG > 1)
		sprintf (sdk2x_private_string, "Partition Table read incorrect! Key = 0x%04X\r\n",*(unsigned short*)&(SdPartInternal.pPartitionTable[510]));
		sdk2x_UartSendString(0,sdk2x_private_string);
#endif
		return SDK2X_ERR_SDPT_PTABLE_CORRUPT;
	}

#if (SDK2X_SD_DEBUG > 1)
	sdk2x_UartSendString(0,"\r\nPartition Table\r\n\r\n");
#endif

	for (i=0;i<4;i++)
	{
#if (SDK2X_SD_DEBUG > 0)
		Partitions[i].nDriveNum = GetInt (SdPartInternal.pPartitionTable,446 + (i<<4),1);
		Partitions[i].nStartCylinder = GetInt (SdPartInternal.pPartitionTable,447 + (i<<4),1) +
				((GetInt (SdPartInternal.pPartitionTable,449,1)&0xc0)<<2);
		Partitions[i].nStartHead = GetInt (SdPartInternal.pPartitionTable,448 + (i<<4),1);
		Partitions[i].nStartSector = GetInt (SdPartInternal.pPartitionTable,449 + (i<<4),1) & 0x3f;
		Partitions[i].nEndCylinder = GetInt (SdPartInternal.pPartitionTable,451 + (i<<4),1) +
				((GetInt (SdPartInternal.pPartitionTable,453 + (i<<4),1)&0xc0)<<2);
		Partitions[i].nEndHead = GetInt (SdPartInternal.pPartitionTable,452 + (i<<4),1);
		Partitions[i].nEndSector = GetInt (SdPartInternal.pPartitionTable,453 + (i<<4),1) & 0x3f;
#endif
		Partitions[i].nType = GetInt (SdPartInternal.pPartitionTable,450 + (i<<4),1);
		Partitions[i].nStartLBA = GetInt (SdPartInternal.pPartitionTable,454 + (i<<4), 4);
		Partitions[i].nSize = GetInt (SdPartInternal.pPartitionTable,458 + (i<<4),4);

#if (SDK2X_SD_DEBUG > 1)
		if (Partitions[i].nType)
		{
			sprintf (sdk2x_private_string, "Partition %d: Start CHS = %4d %3d %2d End CHS = %4d %3d %2d Type = %02X\r\n",
						i, Partitions[i].nStartCylinder, Partitions[i].nStartHead, Partitions[i].nStartSector,
						Partitions[i].nEndCylinder, Partitions[i].nEndHead, Partitions[i].nEndSector,
						Partitions[i].nType);
			sdk2x_UartSendString(0,sdk2x_private_string);
			sprintf (sdk2x_private_string,"             DriveNum = %02X Start LBA = %d Size = %d\r\n",
						Partitions[i].nDriveNum, Partitions[i].nStartLBA, Partitions[i].nSize);
			sdk2x_UartSendString(0,sdk2x_private_string);
		}
#endif
	}

#if (SDK2X_SD_DEBUG > 1)
	sdk2x_UartSendString(0,"\r\n");
#endif
	/* TODO: Validate partition entries to make sure they are within the physical device limits,
				don't overlap etc. */

	for (i=0; i<4; i++)
	{
		if (Partitions[i].nType)
		{
			r = LoadPartitionBootRecord (i);
			if (r)
			{
				bad_partition = i;
				return r;
			}
		}
	}

	return 0;
}

